from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import post_delete
from django.dispatch import receiver

from ..constants import VERIFIED
from ..models import BoxItem
from .box import Box
from .manifest import ManifestItem


@receiver(
    post_delete,
    weak=False,
    sender=ManifestItem,
    dispatch_uid="manifest_item_on_post_delete",
)
def manifest_item_on_post_delete(sender, instance, using, **kwargs):
    try:
        box = Box.objects.get(box_identifier=instance.identifier)
    except ObjectDoesNotExist:
        pass
    else:
        box.status = VERIFIED
        box.save()


@receiver(post_delete, weak=False, sender=BoxItem, dispatch_uid="box_item_on_post_delete")
def box_item_on_post_delete(sender, instance, using, **kwargs):
    instance.box.save()



from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase

from edc_lab.lab import (
    AliquotType,
    LabProfile,
    ProcessingProfile,
    RequisitionPanel,
    RequisitionPanelLookupError,
)
from edc_lab.models import Panel
from edc_lab.site_labs import site_labs


class TestPanel(TestCase):
    def setUp(self):
        site_labs._registry = {}

    def test_panel(self):
        Panel.objects.create(
            name="panel", display_name="My Panel", lab_profile_name="lab_profile"
        )

    def test_panel2(self):

        wb = AliquotType(name="Whole Blood", alpha_code="WB", numeric_code="02")

        whole_blood_processing = ProcessingProfile(name="whole_blood_store", aliquot_type=wb)

        wb_panel = RequisitionPanel(
            name="wb_storage",
            verbose_name="Whole Blood Storage",
            processing_profile=whole_blood_processing,
        )

        lab_profile = LabProfile(
            name="test_profile", requisition_model="edc_lab.subjectrequisition"
        )
        lab_profile.add_panel(wb_panel)

        site_labs.register(lab_profile=lab_profile)

        try:
            Panel.objects.get(name="wb_storage")
        except ObjectDoesNotExist:
            self.fail("Panel unexpectedly does not exist")

        panel = Panel.objects.get(name="wb_storage")

        self.assertEqual(panel.display_name, "Whole Blood Storage")
        self.assertEqual(panel.lab_profile_name, "test_profile")

    def test_requisition_panel_str(self):
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        panel = RequisitionPanel(name="Viral Load", processing_profile=processing_profile)
        self.assertTrue(str(panel))

    def test_requisition_panel(self):
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        RequisitionPanel(name="Viral Load", processing_profile=processing_profile)

    def test_requisition_panel_does_not_know_requisition_model(self):
        """Demonstrate that a panel not yet added to a lab profile
        does not know the requisition model.
        """
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        panel = RequisitionPanel(name="Viral Load", processing_profile=processing_profile)
        self.assertIsNone(panel.requisition_model)
        self.assertRaises(RequisitionPanelLookupError, getattr, panel, "requisition_model_cls")

    def test_requisition_panel_raises_on_invalid_requisition_model(self):
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        for requisition_model in [None, "edc_lab.blah", "blah"]:
            with self.subTest(requisition_model=requisition_model):
                panel = RequisitionPanel(
                    name="Viral Load", processing_profile=processing_profile
                )
                # manually set, normally done by LabProfile
                panel.requisition_model = requisition_model
                try:
                    panel.requisition_model_cls
                except RequisitionPanelLookupError:
                    pass
                else:
                    self.fail("RequisitionPanelModelError unexpectedly not raised.")



# Generated by Django 2.0.7 on 2018-07-06 13:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0015_manifestitem_site")]

    operations = [
        migrations.AlterModelOptions(
            name="historicalaliquot",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical aliquot",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalbox",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical box",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalboxitem",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical box item",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalconsignee",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical consignee",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmanifest",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical manifest",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalorder",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical order",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalresult",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical result",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalresultitem",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical result item",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalshipper",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical shipper",
            },
        ),
    ]



import re


class AliquotTypeAlphaCodeError(Exception):
    pass


class AliquotTypeNumericCodeError(Exception):
    pass


class AliquotType:

    """A class to represent an aliquot type by an alpha and
    numeric code.

    An aliquot type manages a list of valid derivatives for the
    aliquot type, e.g. WB->Plasma, WB->Buffy Coat.
    """

    def __init__(self, name: str = None, alpha_code: str = None, numeric_code: str = None):
        self.derivatives = []
        self.name = name
        if not alpha_code or not re.match(r"^[A-Z]+$", alpha_code, re.ASCII):
            raise AliquotTypeAlphaCodeError(f"Invalid alpha code. Got {alpha_code}.")
        else:
            self.alpha_code = alpha_code
        if not numeric_code or not re.match(r"^\d+$", numeric_code, re.ASCII):
            raise AliquotTypeNumericCodeError(f"Invalid numeric code. Got {numeric_code}.")
        else:
            self.numeric_code = numeric_code

    def __repr__(self):
        return "{self.__class__.__name__}({self.name}, {self.alpha_code}, {self.numeric_code})"

    def __str__(self):
        alpha_code = self.alpha_code or "?alpha_code"
        numeric_code = self.numeric_code or "?numeric_code"
        return f"{self.name.title()} ({alpha_code}:{numeric_code})"

    def add_derivatives(self, *aliquot_type: "AliquotType") -> None:
        """Adds an aliquot instance that is a valid
        derivative of self.
        """
        self.derivatives.extend(aliquot_type)



from .aliquot_type import AliquotType


class ProcessingProfileInvalidDerivative(Exception):
    pass


class ProcessingProfileAlreadyAdded(Exception):
    pass


class Process:

    """A class to represent the resulting aliquot type and number of
    aliquots from the processing of a source aliquot.
    """

    def __init__(self, aliquot_type=None, aliquot_count=None):
        self.aliquot_type = aliquot_type
        self.aliquot_count = aliquot_count
        self.name = f"{self.aliquot_type.name} x {self.aliquot_count}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.aliquot_type}, {self.aliquot_count})"

    def __str__(self):
        return f"Process {self.name}"


class ProcessingProfile:

    """A container of process instances.

    Given a source aliquot, all processes in the profile
    will be "performed" to result in new aliquots of types
    and counts as per the processes.

    Only processes that produce aliquot types that match
    a `derivative` of the profiles aliquot type are accepted.
    """

    process_cls = Process

    def __init__(
        self, name: str = None, aliquot_type: AliquotType = None, verbose_name: str = None
    ):
        self.aliquot_type = aliquot_type
        self.name = name
        self.processes = {}
        self.verbose_name = verbose_name or " ".join(name.split("_")).title()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.aliquot_type})"

    def __str__(self):
        return f"Processing profile {self.verbose_name}"

    def add_processes(self, *processes: Process) -> None:
        """Adds processes to the processing profile or raises."""
        for process in processes:
            if process.aliquot_type not in self.aliquot_type.derivatives:
                raise ProcessingProfileInvalidDerivative(
                    f"Invalid process for profile. Got '{process}'. "
                    f"'{process.aliquot_type}' cannot be derived "
                    f"from '{self.aliquot_type}'."
                )
            if process.name in self.processes:
                raise ProcessingProfileAlreadyAdded(
                    f"Process '{process.name}' has already been added "
                    f"to this processing profile ('{self.name}')."
                )
            self.processes.update({process.name: process})



from zoneinfo import ZoneInfo

from arrow.arrow import Arrow
from django.apps import apps as django_apps
from django.conf import settings
from edc_label import Label
from edc_protocol import Protocol


class RequisitionLabel(Label):
    template_name = "requisition"
    registered_subject_model = "edc_registration.registeredsubject"

    def __init__(self, pk=None, requisition=None, item=None, **kwargs):  # noqa
        super().__init__(label_template_name=self.template_name)
        self._registered_subject = None
        self.item = item or 1
        if requisition:
            self.requisition = requisition
        else:
            model, pk = pk
            self.requisition = django_apps.get_model(model).objects.get(pk=pk)
        self.label_name = self.requisition.human_readable_identifier

    @property
    def registered_subject(self):
        if not self._registered_subject:
            model_cls = django_apps.get_model(self.registered_subject_model)
            self._registered_subject = model_cls.objects.get(
                subject_identifier=self.requisition.subject_identifier
            )
        return self._registered_subject

    @property
    def label_context(self):
        tz = ZoneInfo(settings.TIME_ZONE)
        local = Arrow.fromdatetime(
            self.requisition.drawn_datetime or self.requisition.created
        ).to(tz)
        formatted_date = local.format("YYYY-MM-DD HH:mm")
        printed = "PRINTED: " if not self.requisition.drawn_datetime else "DRAWN: "
        drawn_datetime = f"{printed}{formatted_date}"
        try:
            clinician_initials = self.requisition.user_created[0:2].upper()
        except IndexError:
            clinician_initials = "??"
        return {
            "requisition_identifier": self.requisition.requisition_identifier,
            "item": self.item,
            "item_count": self.requisition.item_count or 1,
            "primary": "<P>",
            "barcode_value": self.requisition.requisition_identifier,
            "protocol": Protocol().protocol,
            "site": str(self.requisition.site.id),
            "site_name": str(self.requisition.site.name),
            "site_title": str(self.requisition.site.siteprofile.title),
            "clinician_initials": clinician_initials,
            "drawn_datetime": drawn_datetime,
            "subject_identifier": self.registered_subject.subject_identifier,
            "gender": self.registered_subject.gender,
            "dob": self.registered_subject.dob,
            "initials": self.registered_subject.initials,
            "identity": self.registered_subject.identity,
            "alpha_code": self.requisition.panel_object.alpha_code,
            "panel": self.requisition.panel_object.abbreviation,
            "panel_name": self.requisition.panel.display_name,
        }



from .aliquot_form import AliquotForm
from .box_form import BoxForm
from .box_item_form import BoxItemForm
from .box_type_form import BoxTypeForm
from .manifest_form import ManifestForm
from .manifest_item_form import ManifestItemForm
from .requisition_form_mixin import RequisitionFormMixin



from datetime import timedelta

from django import forms
from django.conf import settings
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_constants.constants import NO, NOT_APPLICABLE, OTHER, YES
from edc_facility.import_holidays import import_holidays
from edc_form_validators import FormValidator, FormValidatorMixin
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_lab.form_validators import RequisitionFormValidator
from edc_lab.forms import BoxForm, BoxTypeForm, ManifestForm, RequisitionFormMixin
from edc_lab.models import Aliquot

from ...form_validators.requisition_form_validator import RequisitionFormValidatorMixin
from ..models import SubjectConsent, SubjectRequisition, SubjectVisit
from ..site_labs_test_helper import SiteLabsTestHelper
from ..visit_schedules import visit_schedule


class TestForms(TestCase):
    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        return super().setUpClass()

    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)

    def test_box_form_specimen_types1(self):
        data = {"specimen_types": "12, 13"}
        form = BoxForm(data=data)
        form.is_valid()
        self.assertNotIn("specimen_types", list(form.errors.keys()))

    def test_box_form_specimen_types2(self):
        data = {"specimen_types": None}
        form = BoxForm(data=data)
        form.is_valid()
        self.assertIn("specimen_types", list(form.errors.keys()))

    def test_box_form_specimen_types3(self):
        data = {"specimen_types": "AA, BB"}
        form = BoxForm(data=data)
        form.is_valid()
        self.assertIn("specimen_types", list(form.errors.keys()))

    def test_box_form_specimen_types4(self):
        data = {"specimen_types": "12, 13, AA"}
        form = BoxForm(data=data)
        form.is_valid()
        self.assertIn("specimen_types", list(form.errors.keys()))

    def test_box_form_specimen_types5(self):
        data = {"specimen_types": "12, 13, 13"}
        form = BoxForm(data=data)
        form.is_valid()
        self.assertIn("specimen_types", list(form.errors.keys()))

    def test_box_type_form1(self):
        data = {"across": 5, "down": 6, "total": 30}
        form = BoxTypeForm(data=data)
        form.is_valid()
        self.assertNotIn("total", list(form.errors.keys()))

    def test_box_type_form2(self):
        data = {"across": 5, "down": 6, "total": 10}
        form = BoxTypeForm(data=data)
        form.is_valid()
        self.assertIn("total", list(form.errors.keys()))

    def test_manifest_form1(self):
        data = {"category": OTHER, "category_other": None}
        form = ManifestForm(data=data)
        form.is_valid()
        self.assertIn("category_other", list(form.errors.keys()))

    def test_manifest_form2(self):
        data = {"category": "blah", "category_other": None}
        form = ManifestForm(data=data)
        form.is_valid()
        self.assertNotIn("category_other", list(form.errors.keys()))

    def test_requisition_form_reason(self):
        class MyRequisitionFormValidator(RequisitionFormValidatorMixin, FormValidator):
            report_datetime_field_attr = "requisition_datetime"

            @property
            def report_datetime(self):
                return self.cleaned_data.get(self.report_datetime_field_attr)

        data = {"is_drawn": YES, "reason_not_drawn": NOT_APPLICABLE}
        form_validator = MyRequisitionFormValidator(
            cleaned_data=data, model=SubjectRequisition
        )
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertNotIn("reason_not_drawn", cm.exception.error_dict)

        data = {
            "is_drawn": NO,
            "reason_not_drawn": "collection_failed",
            "item_type": NOT_APPLICABLE,
        }
        form_validator = MyRequisitionFormValidator(
            cleaned_data=data, model=SubjectRequisition
        )
        try:
            form_validator.validate()
        except ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_requisition_form_drawn_not_drawn(self):
        class MyRequisitionFormValidator(RequisitionFormValidatorMixin, FormValidator):
            report_datetime_field_attr = "requisition_datetime"

            @property
            def report_datetime(self):
                return self.cleaned_data.get(self.report_datetime_field_attr)

        data = {"is_drawn": YES, "drawn_datetime": None, "requisition_datetime": get_utcnow()}
        form_validator = MyRequisitionFormValidator(
            cleaned_data=data, model=SubjectRequisition
        )
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("drawn_datetime", cm.exception.error_dict)

        self.assertEqual(
            cm.exception.error_dict.get("drawn_datetime")[0].message,
            "This field is required.",
        )

        data = {"is_drawn": NO, "drawn_datetime": get_utcnow()}
        form_validator = MyRequisitionFormValidator(
            cleaned_data=data, model=SubjectRequisition
        )
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("drawn_datetime", cm.exception.error_dict)
        self.assertEqual(
            cm.exception.error_dict.get("drawn_datetime")[0].message,
            "This field is not required.",
        )

        data = {"is_drawn": NO, "drawn_datetime": None}
        form_validator = MyRequisitionFormValidator(
            cleaned_data=data, model=SubjectRequisition
        )
        try:
            form_validator.validate()
        except ValidationError:
            self.fail("ValidationError unexpectedly raised")


class TestForms2(TestCase):

    lab_helper = SiteLabsTestHelper()

    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        self.lab_helper.setup_site_labs()

        class RequisitionForm(RequisitionFormMixin, FormValidatorMixin, forms.ModelForm):

            form_validator_cls = RequisitionFormValidator

            class Meta:
                fields = "__all__"
                model = SubjectRequisition

        self.form_cls = RequisitionForm
        self.subject_identifier = "12345"
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow(),
            identity="1111111",
            confirm_identity="1111111",
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
        )
        appointment = Appointment.objects.get(visit_code="1000")
        self.subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )

    def test_requisition_form_packed_cannot_change(self):
        obj = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.lab_helper.panel.panel_model_obj,
            packed=True,
            processed=True,
            received=True,
        )
        data = {"packed": False, "processed": True, "received": True}
        form = self.form_cls(data=data, instance=obj)
        form.is_valid()
        self.assertIn("packed", list(form.errors.keys()))

    def test_requisition_form_processed_can_change_if_no_aliquots(self):
        obj = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.lab_helper.panel.panel_model_obj,
            packed=True,
            processed=True,
            received=True,
        )
        data = {"packed": True, "processed": False, "received": True}
        form = self.form_cls(data=data, instance=obj)
        form.is_valid()
        self.assertNotIn("processed", list(form.errors.keys()))

    def test_requisition_form_processed_cannot_change_if_aliquots(self):
        obj = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.lab_helper.panel.panel_model_obj,
            packed=True,
            processed=True,
            received=True,
        )
        Aliquot.objects.create(
            aliquot_identifier="1111",
            requisition_identifier=obj.requisition_identifier,
            count=1,
        )
        data = {"packed": True, "processed": False, "received": True}
        form = self.form_cls(data=data, instance=obj)
        form.is_valid()
        self.assertIn("processed", list(form.errors.keys()))

    def test_requisition_form_received_cannot_change(self):
        obj = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.lab_helper.panel.panel_model_obj,
            packed=True,
            processed=True,
            received=True,
        )
        data = {"packed": True, "processed": True, "received": False}
        form = self.form_cls(data=data, instance=obj)
        form.is_valid()
        self.assertIn("received", list(form.errors.keys()))

    def test_requisition_form_received_cannot_be_set_by_form(self):
        obj = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.lab_helper.panel.panel_model_obj,
            received=False,
        )
        data = {"received": True}
        form = self.form_cls(data=data, instance=obj)
        form.is_valid()
        self.assertIn("received", list(form.errors.keys()))

    def test_requisition_form_cannot_be_changed_if_received(self):
        obj = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.lab_helper.panel.panel_model_obj,
            received=True,
        )
        data = {"received": True}
        form = self.form_cls(data=data, instance=obj)
        form.is_valid()
        self.assertIn(
            "Requisition may not be changed", "".join(form.errors.get(NON_FIELD_ERRORS))
        )

    def test_requisition_form_dates(self):
        class RequisitionForm(RequisitionFormMixin, FormValidatorMixin, forms.ModelForm):

            form_validator_cls = RequisitionFormValidator

            class Meta:
                fields = "__all__"
                model = SubjectRequisition

        data = {
            "is_drawn": YES,
            "drawn_datetime": self.subject_visit.report_datetime,
            "requisition_datetime": self.subject_visit.report_datetime - timedelta(days=3),
            "subject_visit": self.subject_visit.pk,
            "report_datetime": self.subject_visit.report_datetime - timedelta(days=3),
        }
        form = RequisitionForm(data=data)
        form.is_valid()
        self.assertIn("requisition_datetime", form._errors)
        self.assertIn(
            "Invalid. Expected a date/time between", form.errors.get("requisition_datetime")[0]
        )



from django.apps import apps as django_apps
from edc_label import Label


class BaseLabel(Label):

    model = None  # lower_label
    template_name = None

    def __init__(self, pk=None, request=None, template_name=None, **kwargs):
        super().__init__(label_template_name=template_name or self.template_name)
        self.request = request
        self.model_cls = django_apps.get_model(self.model)
        self.model_obj = self.model_cls.objects.get(pk=pk)
        self.label_name = self.model_obj.human_readable_identifier



from typing import Protocol


class AliquotTypeStub(Protocol):
    derivatives: list
    name: str
    alpha_code: str
    numeric_code: str

    def add_derivatives(self: "AliquotTypeStub") -> None:
        ...
